﻿// <copyright file="GridViewSync.cs" company="Selectron Systems AG">
// Copyright (c) Selectron Systems AG. All rights reserved.
// </copyright>

namespace ObjectDiffer.Presentation.Evaluation.WpfExtensions
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.Linq;
    using System.Windows;
    using System.Windows.Controls;
    using System.Windows.Controls.Primitives;
    using System.Windows.Media;

    using Telerik.Windows.Controls;
    using Telerik.Windows.Controls.GridView;
    using Telerik.Windows.Controls.TreeListView;

    /// <summary>
    /// Extensions for <see cref="RadGridView"/>
    /// </summary>
    public static class GridViewSync
    {
        /// <summary>
        /// Attached property which contains the Synchronization Group name
        /// </summary>
        public static readonly DependencyProperty SyncGroupProperty = DependencyProperty.RegisterAttached(
            "SyncGroup", typeof(string), typeof(GridViewDataControl), new PropertyMetadata(OnSyncGroupChanged));

        /// <summary>
        /// Attached property containing whether GridView ScrollBars are synchronized or not.
        /// </summary>
        public static readonly DependencyProperty ScrollSyncProperty = DependencyProperty.RegisterAttached(
            "ScrollSync", typeof(bool), typeof(GridViewDataControl), new PropertyMetadata(false));

        /// <summary>
        /// Attached property containing whether GridView Rows height are synchronized or not.
        /// </summary>
        public static readonly DependencyProperty RowHeightSyncProperty = DependencyProperty.RegisterAttached(
            "RowHeightSync", typeof(bool), typeof(GridViewDataControl), new PropertyMetadata(false));

        /// <summary>
        /// Attached property containing whether GridView Rows selections are synchronized or not.
        /// </summary>
        public static readonly DependencyProperty SelectionSyncProperty = DependencyProperty.RegisterAttached(
            "SelectionSync", typeof(bool), typeof(GridViewDataControl), new PropertyMetadata(false));

        private const double Epsilon = 1;

        /// <summary>
        /// List of all registered GridViews by Group name.
        /// </summary>
        private static readonly Dictionary<GridViewDataControl, string> GridViews = new Dictionary<GridViewDataControl, string>();

        /// <summary>
        /// List of all registered GridView ScrollViewers by Group name
        /// </summary>
        private static readonly Dictionary<GridViewScrollViewer, string> ScrollViewers = new Dictionary<GridViewScrollViewer, string>();

        /// <summary>
        /// List of all registered Vertical ScrollBars by Group name
        /// </summary>
        private static readonly Dictionary<ScrollBar, string> VerticalScrollBars = new Dictionary<ScrollBar, string>();

        /// <summary>
        /// List of all registered Horizontal ScrollBars by Group name
        /// </summary>
        private static readonly Dictionary<ScrollBar, string> HorizontalScrollBars = new Dictionary<ScrollBar, string>();

        /// <summary>
        /// Gets the synchronize group.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <returns>The Sync Group.</returns>
        public static string GetSyncGroup(DependencyObject obj)
        {
            return (string)obj.GetValue(SyncGroupProperty);
        }

        /// <summary>
        /// Sets the synchronize group.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <param name="syncGroup">The synchronize group.</param>
        public static void SetSyncGroup(DependencyObject obj, string syncGroup)
        {
            obj.SetValue(SyncGroupProperty, syncGroup);
        }

        /// <summary>
        /// Gets the scroll synchronize.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <returns>Is the scroll synchronized.</returns>
        public static bool GetScrollSync(DependencyObject obj)
        {
            return (bool)obj.GetValue(ScrollSyncProperty);
        }

        /// <summary>
        /// Sets the scroll synchronize.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <param name="scrollSync">The scroll synchronize.</param>
        public static void SetScrollSync(DependencyObject obj, bool scrollSync)
        {
            obj.SetValue(ScrollSyncProperty, scrollSync);
        }

        /// <summary>
        /// Gets the row height synchronize.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <returns>The row height synchronize</returns>
        public static bool GetRowHeightSync(DependencyObject obj)
        {
            return (bool)obj.GetValue(RowHeightSyncProperty);
        }

        /// <summary>
        /// Sets the scroll synchronize.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <param name="rowHeightSync">if set to <c>true</c> [row height synchronize].</param>
        public static void SetRowHeightSync(DependencyObject obj, bool rowHeightSync)
        {
            obj.SetValue(RowHeightSyncProperty, rowHeightSync);
        }

        /// <summary>
        /// Gets the selection synchronize.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <returns>Is the selection synchronized.</returns>
        public static bool GetSelectionSync(DependencyObject obj)
        {
            return (bool)obj.GetValue(SelectionSyncProperty);
        }

        /// <summary>
        /// Sets the selection synchronize.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <param name="selectionSync">The selection synchronize.</param>
        public static void SetSelectionSync(DependencyObject obj, bool selectionSync)
        {
            obj.SetValue(SelectionSyncProperty, selectionSync);
        }

        /// <summary>
        /// Called when [synchronize group changed].
        /// </summary>
        /// <param name="d">The d.</param>
        /// <param name="e">The <see cref="DependencyPropertyChangedEventArgs"/> instance containing the event data.</param>
        private static void OnSyncGroupChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            var gridview = d as GridViewDataControl;
            if (gridview != null)
            {
                // if the Synchronization Group name has changed, remove the GridView from the registered list and remove event handler.
                if (!string.IsNullOrEmpty((string)e.OldValue))
                {
                    // Remove GridView from registered GridViews
                    if (GridViews.ContainsKey(gridview))
                    {
                        ClearEvents(gridview);
                        GridViews.Remove(gridview);
                    }
                }

                if (!string.IsNullOrEmpty((string)e.NewValue))
                {
                    // Add GridView to registered GridViews and attach Load & Unload events. 
                    // Load event is required to be able to find child controls (Scrollbars, ScrollViewer ...)
                    gridview.Loaded += OnGridViewLoaded;
                    gridview.Unloaded += OnGridViewUnloaded;
                    gridview.SelectionChanged += OnGridViewSelectionChanged;
                    gridview.RowIsExpandedChanged += OnGridViewRowIsExpandedChanged;
                    GridViews.Add(gridview, (string)e.NewValue);
                }
            }
        }

        /// <summary>
        /// Called when [grid view load].
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private static void OnGridViewLoaded(object sender, RoutedEventArgs e)
        {
            // ScrollViewer and scrollBars events are registered in the GridView Load event,
            // because we don't have access to the child controls before.
            var gridView = sender as GridViewDataControl;
            if (gridView != null)
            {
                // Get GridView sync group and child controls
                var group = GridViews[gridView];
                var scrollViewer = gridView.ChildrenOfType<GridViewScrollViewer>().FirstOrDefault();
                var verticalScrollBar = gridView.ChildrenOfType<ScrollBar>().FirstOrDefault(sb => sb.Name.Equals("PART_VerticalScrollBar")); // This control name comes from Telerik default control template
                var horizontalScrollBar = gridView.ChildrenOfType<ScrollBar>().FirstOrDefault(sb => sb.Name.Equals("PART_HorizontalScrollBar")); // This control name comes from Telerik default control template

                // Enable Scrollbars synchronization
                var isScrollSync = GetScrollSync(gridView);
                if (isScrollSync)
                {
                    // if Vertical ScrollBar is found, registered it with the group name and attach ValueChanged event
                    if (verticalScrollBar != null)
                    {
                        verticalScrollBar.ValueChanged += OnVerticalScrollBarValueChanged;
                        VerticalScrollBars.Add(verticalScrollBar, group);
                    }

                    // if Horizontal ScrollBar is found, registered it with the group name and attach ValueChanged event
                    if (horizontalScrollBar != null)
                    {
                        horizontalScrollBar.ValueChanged += OnHorizontalScrollBarValueChanged;
                        HorizontalScrollBars.Add(horizontalScrollBar, group);
                    }
                }

                // Enable Row Height synchronization
                var isRowHeightSync = GetRowHeightSync(gridView);
                if (isRowHeightSync && scrollViewer != null)
                {
                    scrollViewer.Loaded += OnGridViewScrollViewerLoaded;
                    ScrollViewers.Add(scrollViewer, group);
                }
            }
        }

        /// <summary>
        /// Called when [grid view unloaded].
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private static void OnGridViewUnloaded(object sender, RoutedEventArgs e)
        {
            var gridView = sender as GridViewDataControl;

            // If the gridView in registered in the registered GridViews, ensure all events are unregistered and remove it from the registered list
            if (gridView != null && GridViews.ContainsKey(gridView))
            {
                ClearEvents(gridView);
                GridViews.Remove(gridView);
            }
        }

        /// <summary>
        /// Called when [horizontal scroll bar value changed].
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The e.</param>
        private static void OnHorizontalScrollBarValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            var scrollBar = (ScrollBar)sender;
            if (scrollBar != null)
            {
                // Retrieve other registered Horizontal Scrollbars in the group
                var group = HorizontalScrollBars[scrollBar];
                var otherScrollBars = HorizontalScrollBars.Where(d => d.Value.Equals(group, StringComparison.OrdinalIgnoreCase) && !d.Key.Equals(scrollBar))
                                                          .Select(d => d.Key)
                                                          .ToList();

                // Manually scroll to horizontal offset for all other scrollbars in the group
                if (otherScrollBars.Any())
                {
                    foreach (var sb in otherScrollBars)
                    {
                        sb.ParentOfType<GridViewScrollViewer>().ScrollToHorizontalOffset(e.NewValue);
                    }
                }
            }
        }

        /// <summary>
        /// Called when [vertical scroll bar value changed].
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The e.</param>
        private static void OnVerticalScrollBarValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            var scrollBar = (ScrollBar)sender;
            if (scrollBar != null)
            {
                // Retrieve other registered Vertical Scrollbars in the group
                var group = VerticalScrollBars[scrollBar];
                var otherScrollBars = VerticalScrollBars.Where(d => d.Value.Equals(group, StringComparison.OrdinalIgnoreCase) && !d.Key.Equals(scrollBar))
                                                        .Select(d => d.Key).Where(sb => Math.Abs(sb.Value - e.NewValue) > Epsilon)
                                                        .ToList();
                if (otherScrollBars.Any())
                {
                    // Manually scroll to vertical offset for all other scrollbars in the group
                    foreach (var sb in otherScrollBars)
                    {
                        var scrollViewer = sb.ParentOfType<GridViewScrollViewer>();
                        scrollViewer.ScrollToVerticalOffset(e.NewValue);
                    }
                }
                else
                {
                    // All scrollbars in the group have been synchronized at this point!
                    // we can retrieve the parent scrollbar, and check if the row height synchronization is enabled.
                    var gridview = scrollBar.ParentOfType<GridViewDataControl>();

                    // if row heigth synchronization is enabled
                    if (gridview != null && GetRowHeightSync(gridview))
                    {
                        // retrieve all ScrollViewers registered in the group
                        var scrollViewers = ScrollViewers.Where(x => x.Value.Equals(group, StringComparison.OrdinalIgnoreCase)).Select(x => x.Key).ToList();

                        // Apply Rows height synchronization
                        SynchronizeRowsHeight(scrollViewers);
                    }
                }
            }
        }

        /// <summary>
        /// Called when [grid view selection changed].
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="SelectionChangeEventArgs"/> instance containing the event data.</param>
        private static void OnGridViewSelectionChanged(object sender, SelectionChangeEventArgs e)
        {
            var currentGridView = sender as GridViewDataControl;

            // Ensure the GridView Selection Synchronization is enabled
            if (!GetSelectionSync(currentGridView))
            {
                return;
            }

            // if a rows has been selected, we know the DataContext of it
            if (currentGridView != null && currentGridView.Items != null && currentGridView.Items.Count > 0 && e.AddedItems.Count == 1)
            {
                // find the index in of the selected row in the GridView data source
                var selectedItemIndex = currentGridView.Items.IndexOf(e.AddedItems[0]);
                if (selectedItemIndex == -1)
                {
                    return;
                }

                // Retrieve other registered GridViews in the group
                var group = GridViews[currentGridView];
                var gridViewsInGroup = GridViews.Where(gv => gv.Value.Equals(group, StringComparison.OrdinalIgnoreCase) && !gv.Key.Equals(currentGridView))
                                                .Select(gv => gv.Key)
                                                .ToList();

                // Iterate through other GridViews
                foreach (var gv in gridViewsInGroup)
                {
                    if (gv.Items != null && gv.Items.Count > 0)
                    {
                        // Find the row at the previously found index, and set this row as selected
                        var itemsToSelect = gv.Items[selectedItemIndex];
                        gv.SelectedItem = itemsToSelect;
                    }
                }
            }
        }

        /// <summary>
        /// Called when [grid view row is expanded changed].
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="RowEventArgs"/> instance containing the event data.</param>
        private static void OnGridViewRowIsExpandedChanged(object sender, RowEventArgs e)
        {
            var currentTreeListView = sender as RadTreeListView;
            var currentRow = (TreeListViewRow)e.Row;

            // Ensure the sender control is a RadTreeListView & current row is a RadTreeListViewRow
            if (currentTreeListView == null || currentRow == null)
            {
                return;
            }

            // Ensure the GridView Selection Synchronization is enabled
            if (!GetSelectionSync(currentTreeListView))
            {
                return;
            }

            // if a rows has been selected, we know the DataContext of it
            if (currentTreeListView.Items != null && currentTreeListView.Items.Count > 0)
            {
                // find the index in of the expanded/collapsed row in the GridView data source
                var toggledItemIndex = currentTreeListView.Items.IndexOf(currentRow.DataContext);
                if (toggledItemIndex == -1)
                {
                    return;
                }

                // Retrieve other registered RadTreeListViews in the group
                var group = GridViews[currentTreeListView];
                var treeListViewsInGroup =
                    GridViews.Where(gv => gv.Value.Equals(group, StringComparison.OrdinalIgnoreCase) && !gv.Key.Equals(currentTreeListView))
                             .Select(gv => gv.Key)
                             .OfType<RadTreeListView>()
                             .ToList();

                // Iterate through other GridViews
                foreach (var tlv in treeListViewsInGroup)
                {
                    if (tlv.Items != null && tlv.Items.Count > 0)
                    {
                        // Find the row at the previously found index, and set this row as selected
                        var itemDataContextToToggle = tlv.Items[toggledItemIndex];
                        var itemToToggle = tlv.ChildrenOfType<TreeListViewRow>().FirstOrDefault(r => r.DataContext == itemDataContextToToggle);
                        if (itemToToggle != null)
                        {
                            itemToToggle.IsExpanded = currentRow.IsExpanded;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Called when [grid view scroll viewer loaded].
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private static void OnGridViewScrollViewerLoaded(object sender, RoutedEventArgs e)
        {
            // This event is used when the GridViewRow height synchronization is enabled.
            // It is used to synchronize the height of all visible rows (Because of virtualization).
            // We have to do this 'first' synchronization here, because not scroll event has been fired yet.
            var scrollViewer = sender as GridViewScrollViewer;
            if (scrollViewer != null)
            {
                var gridView = scrollViewer.ParentOfType<GridViewDataControl>();
                if (gridView != null)
                {
                    // Ensure the GridViewRow height synchronization is enabled
                    if (!GetRowHeightSync(gridView))
                    {
                        return;
                    }

                    var group = GridViews[gridView];
                    if (!string.IsNullOrEmpty(group))
                    {
                        // Retrieve all registered ScrollViewers in the group.
                        var grids = GridViews.Where(d => d.Value.Equals(group, StringComparison.OrdinalIgnoreCase)).Select(d => d.Key).ToList();
                        var scrollViewers = grids.Select(g => g.ChildrenOfType<GridViewScrollViewer>().FirstOrDefault());

                        // Apply Rows height synchronization
                        SynchronizeRowsHeight(scrollViewers);
                    }
                }
            }
        }

        /// <summary>
        /// Synchronizes the height of the rows.
        /// </summary>
        /// <param name="scrollViewers">The scroll viewers.</param>
        private static void SynchronizeRowsHeight(IEnumerable<GridViewScrollViewer> scrollViewers)
        {
            // Retrieve visibles rows by ScrollViewer
            var rowsByScrollViewers = scrollViewers.Select(sv => sv.ChildrenOfType<GridViewRow>().ToList()).ToList();

            // We can do a synchronization only if we have at least two visible ScrollViewers
            if (rowsByScrollViewers.Count >= 2)
            {
                // get the mininum amount of row in the ScrollViewers. This is to avoid IndexOutOfRangeException in case of row number mismatch
                var minRowsCount = rowsByScrollViewers.Select(rows => rows.Count).Min();

                // loop through rows within ScrollViewers
                for (int index = 0; index < minRowsCount; index++)
                {
                    int rowIndex = index;

                    // retrieve rows by ScrollViewer at the given index.
                    var rowsAtIndex = rowsByScrollViewers.Select(col => col.ElementAt(rowIndex)).ToList();

                    // Apply the synchronization only if at least two rows have been found for this index (meaning at least 2 GridViews has a visible rows with this index)
                    if (rowsAtIndex.Count > 1)
                    {
                        // retrieve the height of the highest row (increase by 5, to ensure the padding)
                        var maxHeight = rowsAtIndex.Select(GetRowHeight).Max() + 5;

                        // apply the maximum height to every rows
                        rowsAtIndex.ForEach(r => r.Height = maxHeight);
                    }
                }
            }
        }

        /// <summary>
        /// Unregisters events from the GridView.
        /// </summary>
        /// <param name="gridview">The GridView.</param>
        private static void ClearEvents(GridViewDataControl gridview)
        {
            var scrollViewer = gridview.ChildrenOfType<GridViewScrollViewer>().FirstOrDefault();
            var verticalScrollBar = gridview.ChildrenOfType<ScrollBar>().FirstOrDefault(sb => sb.Name.Equals("PART_VerticalScrollBar"));
            var horizontalScrollBar = gridview.ChildrenOfType<ScrollBar>().FirstOrDefault(sb => sb.Name.Equals("PART_HorizontalScrollBar"));

            if (scrollViewer != null && ScrollViewers.ContainsKey(scrollViewer))
            {
                scrollViewer.Loaded -= OnGridViewScrollViewerLoaded;
                ScrollViewers.Remove(scrollViewer);
            }

            if (verticalScrollBar != null && VerticalScrollBars.ContainsKey(verticalScrollBar))
            {
                verticalScrollBar.ValueChanged -= OnVerticalScrollBarValueChanged;
                VerticalScrollBars.Remove(verticalScrollBar);
            }

            if (horizontalScrollBar != null && HorizontalScrollBars.ContainsKey(horizontalScrollBar))
            {
                horizontalScrollBar.ValueChanged -= OnVerticalScrollBarValueChanged;
                HorizontalScrollBars.Remove(horizontalScrollBar);
            }

            gridview.Loaded -= OnGridViewLoaded;
            gridview.Unloaded -= OnGridViewUnloaded;
            gridview.SelectionChanged -= OnGridViewSelectionChanged;
            gridview.RowIsExpandedChanged -= OnGridViewRowIsExpandedChanged;
        }

        /// <summary>
        /// Measures the size of the text.
        /// </summary>
        /// <param name="text">The text.</param>
        /// <param name="fontFamily">The font family.</param>
        /// <param name="fontStyle">The font style.</param>
        /// <param name="fontWeight">The font weight.</param>
        /// <param name="fontStretch">The font stretch.</param>
        /// <param name="fontSize">Size of the font.</param>
        /// <returns><see cref="Size"/></returns>
        private static Size MeasureTextSize(string text, FontFamily fontFamily, FontStyle fontStyle, FontWeight fontWeight, FontStretch fontStretch, double fontSize)
        {
            var ft = new FormattedText(
                text, CultureInfo.CurrentCulture, FlowDirection.LeftToRight, new Typeface(fontFamily, fontStyle, fontWeight, fontStretch), fontSize, Brushes.Black);
            return new Size(ft.Width, ft.Height);
        }

        /// <summary>
        /// Gets the height of the row.
        /// </summary>
        /// <param name="row">The row.</param>
        /// <remarks>
        /// This methods will retrieve the maximum row height based only on the TextBlocks present in the row cells.
        /// In case of other controls (i.e. Image, ...)
        /// </remarks>
        /// <returns>The row height.</returns>
        private static double GetRowHeight(GridViewRow row)
        {
            return
                row.ChildrenOfType<TextBlock>()
                   .Select(b => MeasureTextSize(b.Text, row.FontFamily, row.FontStyle, row.FontWeight, row.FontStretch, row.FontSize).Height)
                   .Max();
        }
    }
}
